from dataclasses import dataclass
from typing import Dict, Generator

from httpx import Auth, Request, Response

from beamlit.common.settings import Settings, get_settings

from ..client import AuthenticatedClient
from .apikey import ApiKeyProvider
from .clientcredentials import ClientCredentials
from .credentials import (
    Credentials,
    current_context,
    load_credentials,
    load_credentials_from_settings,
)
from .device_mode import BearerToken


class PublicProvider(Auth):
    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        yield request


@dataclass
class RunClientWithCredentials:
    credentials: Credentials
    workspace: str
    api_url: str = "https://api.beamlit.dev/v0"
    run_url: str = "https://run.beamlit.dev/v0"


def new_client_from_settings(settings: Settings):
    credentials = load_credentials_from_settings(settings)

    client_config = RunClientWithCredentials(
        credentials=credentials,
        workspace=settings.workspace,
    )
    return new_client_with_credentials(client_config)


def new_client():
    context = current_context()
    if context.workspace:
        credentials = load_credentials(context.workspace)
        client_config = RunClientWithCredentials(
            credentials=credentials,
            workspace=context.workspace,
        )
    else:
        settings = get_settings()
        credentials = load_credentials_from_settings(settings)
        client_config = RunClientWithCredentials(
            credentials=credentials,
            workspace=settings.workspace,
        )
    return new_client_with_credentials(client_config)


def new_client_with_credentials(config: RunClientWithCredentials):
    provider: Auth = None
    if config.credentials.api_key:
        provider = ApiKeyProvider(config.credentials, config.workspace)
    elif config.credentials.access_token:
        provider = BearerToken(config.credentials, config.workspace, config.api_url)
    elif config.credentials.client_credentials:
        provider = ClientCredentials(config.credentials, config.workspace, config.api_url)
    else:
        provider = PublicProvider()

    return AuthenticatedClient(base_url=config.api_url, provider=provider)


def get_authentication_headers(settings: Settings) -> Dict[str, str]:
    context = current_context()
    if context.workspace:
        credentials = load_credentials(context.workspace)
    else:
        settings = get_settings()
        credentials = load_credentials_from_settings(settings)

    config = RunClientWithCredentials(
        credentials=credentials,
        workspace=settings.workspace,
    )
    provider = None
    if config.credentials.api_key:
        provider = ApiKeyProvider(config.credentials, config.workspace)
    elif config.credentials.access_token:
        provider = BearerToken(config.credentials, config.workspace, config.api_url)
    elif config.credentials.client_credentials:
        provider = ClientCredentials(config.credentials, config.workspace, config.api_url)

    if provider is None:
        return None
    return provider.get_headers()



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.qps import QPS


T = TypeVar("T", bound="ResourceDeploymentMetricsQueryPerRegionPerCode")


@_attrs_define
class ResourceDeploymentMetricsQueryPerRegionPerCode:
    """Number of requests done on the resource for the model deployment

    Attributes:
        region (Union[Unset, QPS]): Query per second per element, can be per response status code (e.g. 200, 400) or per
            location
    """

    region: Union[Unset, "QPS"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        region: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.region, Unset):
            region = self.region.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if region is not UNSET:
            field_dict["region"] = region

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.qps import QPS

        if not src_dict:
            return None
        d = src_dict.copy()
        _region = d.pop("region", UNSET)
        region: Union[Unset, QPS]
        if isinstance(_region, Unset):
            region = UNSET
        else:
            region = QPS.from_dict(_region)

        resource_deployment_metrics_query_per_region_per_code = cls(
            region=region,
        )

        resource_deployment_metrics_query_per_region_per_code.additional_properties = d
        return resource_deployment_metrics_query_per_region_per_code

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="AuthenticationProviderModel")


@_attrs_define
class AuthenticationProviderModel:
    """Model obtained from an external authentication provider, such as HuggingFace, OpenAI, etc...

    Attributes:
        created_at (Union[Unset, str]): Provider model created at
        downloads (Union[Unset, int]): Provider model downloads
        id (Union[Unset, str]): Provider model ID
        library_name (Union[Unset, str]): Provider model library name
        likes (Union[Unset, int]): Provider model likes
        name (Union[Unset, str]): Provider model name
        pipeline_tag (Union[Unset, str]): Provider model pipeline tag
        private (Union[Unset, bool]): Provider model private
        tags (Union[Unset, list[str]]): Provider model tags
        trending_score (Union[Unset, int]): Provider model trending score
    """

    created_at: Union[Unset, str] = UNSET
    downloads: Union[Unset, int] = UNSET
    id: Union[Unset, str] = UNSET
    library_name: Union[Unset, str] = UNSET
    likes: Union[Unset, int] = UNSET
    name: Union[Unset, str] = UNSET
    pipeline_tag: Union[Unset, str] = UNSET
    private: Union[Unset, bool] = UNSET
    tags: Union[Unset, list[str]] = UNSET
    trending_score: Union[Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        created_at = self.created_at

        downloads = self.downloads

        id = self.id

        library_name = self.library_name

        likes = self.likes

        name = self.name

        pipeline_tag = self.pipeline_tag

        private = self.private

        tags: Union[Unset, list[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        trending_score = self.trending_score

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if downloads is not UNSET:
            field_dict["downloads"] = downloads
        if id is not UNSET:
            field_dict["id"] = id
        if library_name is not UNSET:
            field_dict["library_name"] = library_name
        if likes is not UNSET:
            field_dict["likes"] = likes
        if name is not UNSET:
            field_dict["name"] = name
        if pipeline_tag is not UNSET:
            field_dict["pipeline_tag"] = pipeline_tag
        if private is not UNSET:
            field_dict["private"] = private
        if tags is not UNSET:
            field_dict["tags"] = tags
        if trending_score is not UNSET:
            field_dict["trending_score"] = trending_score

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        created_at = d.pop("created_at", UNSET)

        downloads = d.pop("downloads", UNSET)

        id = d.pop("id", UNSET)

        library_name = d.pop("library_name", UNSET)

        likes = d.pop("likes", UNSET)

        name = d.pop("name", UNSET)

        pipeline_tag = d.pop("pipeline_tag", UNSET)

        private = d.pop("private", UNSET)

        tags = cast(list[str], d.pop("tags", UNSET))

        trending_score = d.pop("trending_score", UNSET)

        authentication_provider_model = cls(
            created_at=created_at,
            downloads=downloads,
            id=id,
            library_name=library_name,
            likes=likes,
            name=name,
            pipeline_tag=pipeline_tag,
            private=private,
            tags=tags,
            trending_score=trending_score,
        )

        authentication_provider_model.additional_properties = d
        return authentication_provider_model

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="StoreFunctionConfiguration")


@_attrs_define
class StoreFunctionConfiguration:
    """Store function configuration used to configure your function directly from beamlit interface

    Attributes:
        description (Union[Unset, str]): Store function configuration description
        if_ (Union[Unset, str]): Conditional rendering for the configuration, example: provider === 'openai'
        name (Union[Unset, str]): Store function configuration name
        required (Union[Unset, bool]): Store function configuration required
        secret (Union[Unset, bool]): Store function configuration secret
    """

    description: Union[Unset, str] = UNSET
    if_: Union[Unset, str] = UNSET
    name: Union[Unset, str] = UNSET
    required: Union[Unset, bool] = UNSET
    secret: Union[Unset, bool] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        description = self.description

        if_ = self.if_

        name = self.name

        required = self.required

        secret = self.secret

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if description is not UNSET:
            field_dict["description"] = description
        if if_ is not UNSET:
            field_dict["if"] = if_
        if name is not UNSET:
            field_dict["name"] = name
        if required is not UNSET:
            field_dict["required"] = required
        if secret is not UNSET:
            field_dict["secret"] = secret

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        description = d.pop("description", UNSET)

        if_ = d.pop("if", UNSET)

        name = d.pop("name", UNSET)

        required = d.pop("required", UNSET)

        secret = d.pop("secret", UNSET)

        store_function_configuration = cls(
            description=description,
            if_=if_,
            name=name,
            required=required,
            secret=secret,
        )

        store_function_configuration.additional_properties = d
        return store_function_configuration

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="DeploymentConfiguration")


@_attrs_define
class DeploymentConfiguration:
    """Function configuration, this is a key value storage. In your function you can retrieve the value with config[key]

    Attributes:
        secret (Union[Unset, bool]): Function configuration secret
        value (Union[Unset, str]): Function configuration value
    """

    secret: Union[Unset, bool] = UNSET
    value: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        secret = self.secret

        value = self.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if secret is not UNSET:
            field_dict["secret"] = secret
        if value is not UNSET:
            field_dict["value"] = value

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        secret = d.pop("secret", UNSET)

        value = d.pop("value", UNSET)

        deployment_configuration = cls(
            secret=secret,
            value=value,
        )

        deployment_configuration.additional_properties = d
        return deployment_configuration

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Tuple

from beamlit.common.settings import Settings, get_settings
from beamlit.models.agent_deployment import AgentDeployment
from beamlit.models.function_deployment import FunctionDeployment
from beamlit.models.function_kit import FunctionKit


def get_titles_name(name: str) -> str:
    return name.title().replace("-", "").replace("_", "")


def generate_kit_function_code(settings: Settings, function: FunctionDeployment, kit: FunctionKit) -> Tuple[str, str]:
    export_code = ""
    code = ""
    for kit in kit:
        fn = FunctionDeployment(
            function=kit.name,
            workspace=settings.workspace,
            parameters=kit.parameters,
            description=kit.description,
        )
        new_code, export = generate_function_code(settings, fn, force_name_in_endpoint=function.function, kit=True)
        code += new_code
        export_code += export
    return code, export_code


def generate_function_code(
    settings: Settings, function: FunctionDeployment, force_name_in_endpoint: str = "", kit: bool = False
) -> Tuple[str, str]:
    name = get_titles_name(function.function)
    if function.parameters and len(function.parameters) > 0:
        args_list = ", ".join(f"{param.name}: str" for param in function.parameters)
        args_list += ", "
    else:
        args_list = ""
    args_schema = ""
    if function.parameters:
        for param in function.parameters:
            args_schema += f'{param.name}: str = Field(description="""{param.description}""")\n    '
    if len(args_schema) == 0:
        args_schema = "pass"

    # TODO: add return direct in function configuration
    return_direct = False
    endpoint_name = force_name_in_endpoint or function.function
    body = "{}"
    if function.parameters:
        body = f'{", ".join(f'"{param.name}": {param.name}' for param in function.parameters)}'
    if kit is True:
        has_name = False
        if function.parameters:
            for param in function.parameters:
                if param.name == "name":
                    has_name = True
                    break
        if not has_name:
            if len(body) > 0:
                body += ", "
            body += f'"name": "{function.function}"'
    return (
        f'''

class Beamlit{name}Input(BaseModel):
    {args_schema}

class Beamlit{name}(BaseTool):
    name: str = "beamlit_{function.function.replace("-", "_")}"
    description: str = """{function.description}"""
    args_schema: Type[BaseModel] = Beamlit{name}Input

    response_format: Literal["content_and_artifact"] = "content_and_artifact"
    return_direct: bool = {return_direct}

    def _run(self, {args_list} run_manager: Optional[CallbackManagerForToolRun] = None) -> Tuple[Union[List[Dict[str, str]], str], Dict]:
        try:
            params = self.metadata.get("params", {{}})
            response = run_client.run("function", "{endpoint_name}", settings.environment, "POST", json={{{body}}})
            if response.status_code >= 400:
                logger.error(f"Failed to run function {name}, {{response.status_code}}::{{response.text}}")
                raise Exception(f"Failed to run function {name}, {{response.status_code}}::{{response.text}}")
            return response.json(), {{}}
        except Exception as e:
            return repr(e), {{}}
''',
        f"Beamlit{get_titles_name(function.function)},",
    )


def generate_chain_code(settings: Settings, agent: AgentDeployment) -> Tuple[str, str]:
    name = get_titles_name(agent.agent)
    # TODO: add return direct in agent configuration
    return_direct = False
    return (
        f'''
class BeamlitChain{name}Input(BaseModel):
    input: str = Field(description='{agent.description}')

class BeamlitChain{name}(BaseTool):
    name: str = "beamlit_chain_{agent.agent.replace("-", "_")}"
    description: str = """{agent.description}"""
    args_schema: Type[BaseModel] = BeamlitChain{name}Input

    response_format: Literal["content_and_artifact"] = "content_and_artifact"
    return_direct: bool = {return_direct}

    def _run(
        self,
        input: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> Tuple[Union[List[Dict[str, str]], str], Dict]:
        try:
            params = self.metadata.get("params", {{}})
            response = run_client.run("agent", "{agent.agent}", settings.environment, "POST", json={{"input": input}})
            if response.status_code >= 400:
                logger.error(f"Failed to run tool {agent.agent}, {{response.status_code}}::{{response.text}}")
                raise Exception(f"Failed to run tool {agent.agent}, {{response.status_code}}::{{response.text}}")
            if response.headers.get("Content-Type") == "application/json":
                return response.json(), {{}}
            else:
                return response.text, {{}}
        except Exception as e:
            return repr(e), {{}}
''',
        f"BeamlitChain{name},",
    )


def generate(destination: str, dry_run: bool = False):
    imports = """from logging import getLogger
from typing import Dict, List, Literal, Optional, Tuple, Type, Union

from langchain_core.callbacks import CallbackManagerForToolRun
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field
from beamlit.authentication import (RunClientWithCredentials,
                                    load_credentials_from_settings,
                                    new_client_with_credentials)
from beamlit.common.settings import get_settings
from beamlit.run import RunClient

logger = getLogger(__name__)
settings = get_settings()
credentials = load_credentials_from_settings(settings)

client_config = RunClientWithCredentials(
    credentials=credentials,
    workspace=settings.workspace,
)
client = new_client_with_credentials(client_config)
run_client = RunClient(client=client)
"""
    settings = get_settings()
    export_code = "\n\nfunctions = ["
    export_chain = "\n\nchains = ["
    code = imports
    if settings.agent.functions and len(settings.agent.functions) > 0:
        for function_config in settings.agent.functions:
            if function_config.kit and len(function_config.kit) > 0:
                new_code, export = generate_kit_function_code(settings, function_config, function_config.kit)
                code += new_code
                export_code += export
            else:
                new_code, export = generate_function_code(settings, function_config)
                code += new_code
                export_code += export
    if settings.agent.chain and len(settings.agent.chain) > 0:
        for agent in settings.agent.chain:
            new_code, export = generate_chain_code(settings, agent)
            code += new_code
            export_chain += export
    if settings.agent.functions and len(settings.agent.functions) > 0:
        export_code = export_code[:-1]
    export_code += "]"
    if settings.agent.chain and len(settings.agent.chain) > 0:
        export_chain = export_chain[:-1]
    export_chain += "]"
    content = code + export_code + export_chain
    if not dry_run:
        with open(destination, "w") as f:
            f.write(content)
    return content



import ast
import importlib
import json
import os
import sys
from dataclasses import dataclass
from logging import getLogger
from typing import Callable, Literal

from beamlit.common.settings import Settings, get_settings, init
from beamlit.models import (AgentChain, AgentDeployment, Flavor,
                            FunctionDeployment, Runtime,
                            StoreFunctionParameter)

sys.path.insert(0, os.getcwd())
sys.path.insert(0, os.path.join(os.getcwd(), "src"))


@dataclass
class Resource:
    type: Literal["agent", "function"]
    module: Callable
    name: str
    decorator: ast.Call
    func: Callable

def get_resources(from_decorator, dir="src") -> list[Resource]:
    """
    Scans through Python files in a directory to find functions decorated with a specific decorator.

    Args:
        from_decorator (str): The name of the decorator to search for
        dir (str): The directory to scan, defaults to "src"

    Returns:
        list[Resource]: List of Resource objects containing information about decorated functions
    """
    resources = []
    logger = getLogger(__name__)

    # Walk through all Python files in resources directory and subdirectories
    for root, _, files in os.walk(dir):
        for file in files:
            if file.endswith(".py"):
                file_path = os.path.join(root, file)
                # Read and compile the file content
                with open(file_path) as f:
                    try:
                        file_content = f.read()
                        # Parse the file content to find decorated resources
                        tree = ast.parse(file_content)

                        # Look for function definitions with decorators
                        for node in ast.walk(tree):
                            if (
                                not isinstance(node, ast.FunctionDef) and not isinstance(node, ast.AsyncFunctionDef)
                            ) or len(node.decorator_list) == 0:
                                continue
                            decorator = node.decorator_list[0]

                            decorator_name = ""
                            if isinstance(decorator, ast.Call):
                                decorator_name = decorator.func.id
                            if isinstance(decorator, ast.Name):
                                decorator_name = decorator.id
                            if decorator_name == from_decorator:
                                # Get the function name and decorator name
                                func_name = node.name

                                # Import the module to get the actual function
                                spec = importlib.util.spec_from_file_location(func_name, file_path)
                                module = importlib.util.module_from_spec(spec)
                                spec.loader.exec_module(module)
                                # Check if kit=True in the decorator arguments

                                # Get the decorated function
                                if hasattr(module, func_name) and isinstance(decorator, ast.Call):

                                    resources.append(
                                        Resource(
                                            type=decorator_name,
                                            module=module,
                                            name=func_name,
                                            func=getattr(module, func_name),
                                            decorator=decorator,
                                        )
                                    )
                    except Exception as e:
                        logger.warning(f"Error processing {file_path}: {e!s}")
    return resources


def get_parameters(resource: Resource) -> list[StoreFunctionParameter]:
    """
    Extracts parameter information from a function's signature and docstring.

    Args:
        resource (Resource): The resource object containing the function to analyze

    Returns:
        list[StoreFunctionParameter]: List of parameter objects with name, type, required status, and description
    """
    parameters = []
    # Get function signature
    import inspect
    sig = inspect.signature(resource.func)
    # Get docstring for parameter descriptions
    docstring = inspect.getdoc(resource.func)
    param_descriptions = {}
    if docstring:
        # Parse docstring for parameter descriptions
        lines = docstring.split('\n')
        for line in lines:
            line = line.strip().lower()
            if line.startswith(':param '):
                # Extract parameter name and description
                param_line = line[7:].split(':', 1)
                if len(param_line) == 2:
                    param_name = param_line[0].strip()
                    param_desc = param_line[1].strip()
                    param_descriptions[param_name] = param_desc
    for name, param in sig.parameters.items():
        # Skip *args and **kwargs parameters
        if param.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):

            continue

        param_type = "string"  # Default type
        type_mapping = {
            'str': 'string',
            'int': 'integer',
            'float': 'number',
            'bool': 'boolean',
            'list': 'array',
            'dict': 'object',
            'none': 'null'
        }
        if param.annotation != inspect.Parameter.empty:
            # Map Python types to OpenAPI types
            if hasattr(param.annotation, "__name__"):
                param_type = param.annotation.__name__.lower()
            else:
                # Handle special types like Union, Optional etc
                param_type = str(param.annotation).lower()
        parameter = StoreFunctionParameter(
            name=name,
            type_=type_mapping.get(param_type, "string"),
            required=param.default == inspect.Parameter.empty,
            description=param_descriptions.get(name, f"Parameter {name}")
        )
        parameters.append(parameter)

    return parameters


def get_description(description: str | None, resource: Resource) -> str:
    """
    Gets the description of a function from either a provided description or the function's docstring.

    Args:
        description (str | None): Optional explicit description
        resource (Resource): The resource object containing the function

    Returns:
        str: The function description
    """
    if description:
        return description
    doc = resource.func.__doc__
    if doc:
        # Split docstring into sections and get only the description part
        doc_lines = doc.split('\n')
        description_lines = []
        for line in doc_lines:
            line = line.strip()
            # Stop when we hit param/return sections
            if line.startswith(':param') or line.startswith(':return'):
                break
            if line:
                description_lines.append(line)
        return ' '.join(description_lines).strip()
    return ""

def get_kwargs(arg: ast.Call) -> dict:
    """
    Extracts keyword arguments from an AST Call node.

    Args:
        arg (ast.Call): The AST Call node to process

    Returns:
        dict: Dictionary of keyword arguments and their values
    """
    kwargs = {}
    for keyword in arg.keywords:
        if isinstance(keyword.value, ast.Constant):
            kwargs[keyword.arg] = keyword.value.value
        elif isinstance(keyword.value, (ast.List, ast.Tuple)):
            kwargs[keyword.arg] = [
                AgentChain(**get_kwargs(elem)) if isinstance(elem, ast.Call) and isinstance(elem.func, ast.Name) and elem.func.id == "AgentChain"
                else elem.value if isinstance(elem, ast.Constant) else elem
                for elem in keyword.value.elts
            ]
        elif isinstance(keyword.value, ast.Dict):
            kwargs[keyword.arg] = {}
            for k, v in zip(keyword.value.keys, keyword.value.values):
                if isinstance(k, ast.Constant) and isinstance(v, ast.Constant):
                    kwargs[keyword.arg][k.value] = v.value
                if isinstance(k, ast.Constant) and isinstance(v, ast.Call):
                    kwargs[keyword.arg][k.value] = get_kwargs(v)
    return kwargs

def get_runtime(type: str, name: str) -> Runtime:
    settings = get_settings()
    registry_url = settings.registry_url.replace("https://", "").replace("http://", "")
    image = f"{registry_url}/{settings.workspace}/{type}s/{name}"
    return Runtime(image=image)

def get_beamlit_deployment_from_resource(resource: Resource) -> AgentDeployment | FunctionDeployment:
    """
    Creates a deployment configuration from a resource.

    Args:
        resource (Resource): The resource to create a deployment for

    Returns:
        AgentDeployment | FunctionDeployment: The deployment configuration
    """
    for arg in resource.decorator.args:
        if isinstance(arg, ast.Call):
            if isinstance(arg.func, ast.Name) and arg.func.id == "AgentDeployment":
                kwargs = get_kwargs(arg)
                description = kwargs.pop("description", None)
                return AgentDeployment(
                    **kwargs,
                    description=get_description(description, resource),
                    runtime=get_runtime("agent", kwargs.get("agent", resource.name))
                )
            if isinstance(arg.func, ast.Name) and arg.func.id == "FunctionDeployment":
                kwargs = get_kwargs(arg)
                description = kwargs.pop("description", None)
                return FunctionDeployment(
                    **kwargs,
                    parameters=get_parameters(resource),
                    description=get_description(description, resource),
                    runtime=get_runtime("function", kwargs.get("function", resource.name))
                )
    for arg in resource.decorator.keywords:
        if isinstance(arg.value, ast.Call):
            if isinstance(arg.value.func, ast.Name) and arg.value.func.id == "AgentDeployment":
                kwargs = get_kwargs(arg.value)
                description = kwargs.pop("description", None)
                return AgentDeployment(
                    **kwargs,
                    description=get_description(description, resource),
                    runtime=get_runtime("agent", kwargs.get("agent", resource.name))
                )
            if isinstance(arg.value.func, ast.Name) and arg.value.func.id == "FunctionDeployment":
                kwargs = get_kwargs(arg.value)
                description = kwargs.pop("description", None)
                return FunctionDeployment(
                    **kwargs,
                    parameters=get_parameters(resource),
                    description=get_description(description, resource),
                    runtime=get_runtime("function", kwargs.get("function", resource.name))
                )
    if resource.type == "agent":
        return AgentDeployment(
            agent=resource.name,
            description=get_description(None,resource),
            runtime=get_runtime("agent", resource.name)
        )
    if resource.type == "function":
        return FunctionDeployment(
            function=resource.name,
            parameters=get_parameters(resource),
            description=get_description(None,resource),
            runtime=get_runtime("function", resource.name)
        )
    return None


def get_flavors(flavors: list[Flavor]) -> str:
    """
    Converts a list of Flavor objects to JSON string.

    Args:
        flavors (list[Flavor]): List of Flavor objects

    Returns:
        str: JSON string representation of flavors
    """
    if not flavors:
        return "[]"
    return json.dumps([flavor.to_dict() for flavor in flavors])

def format_parameters(parameters: list[StoreFunctionParameter]) -> str:
    """
    Formats function parameters into YAML-compatible string.

    Args:
        parameters (list[StoreFunctionParameter]): List of parameter objects

    Returns:
        str: YAML-formatted string of parameters
    """
    if not parameters:
        return "[]"

    formatted = []
    for param in parameters:
        formatted.append(f"""
      - name: {param.name}
        type: {param.type_}
        required: {str(param.required).lower()}
        description: {param.description}""")

    return "\n".join(formatted)

def format_agent_chain(agent_chain: list[AgentChain]) -> str:
    """
    Formats agent chain configuration into YAML-compatible string.

    Args:
        agent_chain (list[AgentChain]): List of agent chain configurations

    Returns:
        str: YAML-formatted string of agent chain
    """
    if not agent_chain:
        return "[]"
    formatted = []

    for agent in agent_chain:
        formatted.append(f"""
      - agent: {agent.name}
        enabled: {agent.enabled}""")
        if agent.description:
            formatted.append(f"        description: {agent.description}")
    return "\n".join(formatted)

def get_agent_yaml(agent: AgentDeployment, functions: list[tuple[Resource, FunctionDeployment]], settings: Settings) -> str:
    """
    Generates YAML configuration for an agent deployment.

    Args:
        agent (AgentDeployment): Agent deployment configuration
        functions (list[tuple[Resource, FunctionDeployment]]): List of associated functions
        settings (Settings): Application settings

    Returns:
        str: YAML configuration string
    """
    template = f"""
apiVersion: beamlit.com/v1alpha1
kind: Agent
metadata:
  name: {agent.agent}
spec:
  display_name: {agent.agent}
  deployments:
  - environment: {settings.environment}
    enabled: true
    policies: [{", ".join(agent.policies or [])}]
    functions: [{", ".join([f"{function.function}" for (_, function) in functions])}]
    agent_chain: {format_agent_chain(agent.agent_chain)}
    model: {agent.model}
    runtime:
      image: {agent.runtime.image}
"""
    if agent.description:
        template += f"""    description: |
      {agent.description}"""
    return template

def get_function_yaml(function: FunctionDeployment, settings: Settings) -> str:
    """
    Generates YAML configuration for a function deployment.

    Args:
        function (FunctionDeployment): Function deployment configuration
        settings (Settings): Application settings

    Returns:
        str: YAML configuration string
    """
    return f"""
apiVersion: beamlit.com/v1alpha1
kind: Function
metadata:
  name: {function.function}
spec:
  display_name: {function.function}
  deployments:
  - environment: {settings.environment}
    enabled: true
    policies: [{", ".join(function.policies or [])}]
    description: |
      {function.description}
    parameters: {format_parameters(function.parameters)}
    runtime:
      image: {function.runtime.image}
"""

def dockerfile(type: Literal["agent", "function"], resource: Resource, deployment: AgentDeployment | FunctionDeployment) -> str:
    """
    Generates Dockerfile content for agent or function deployment.

    Args:
        type (Literal["agent", "function"]): Type of deployment
        resource (Resource): Resource to be deployed
        deployment (AgentDeployment | FunctionDeployment): Deployment configuration

    Returns:
        str: Dockerfile content
    """
    if type == "agent":
        module = f"{resource.module.__file__.split('/')[-1].replace('.py', '')}.{resource.module.__name__}"
    else:
        module = f"functions.{resource.module.__name__}.{resource.func.__name__}"
    cmd = ["bl", "serve", "--port", "80", "--module", module]
    if type == "agent":
        cmd.append("--remote")
    cmd_str = ','.join([f'"{c}"' for c in cmd])

    return f"""
FROM python:3.12-slim

ARG UV_VERSION="latest"
RUN apt update && apt install -y curl

# Install uv.
COPY --from=ghcr.io/astral-sh/uv:latest /uv /uvx /bin/
RUN curl -fsSL https://raw.githubusercontent.com/beamlit/toolkit/main/install.sh | BINDIR=/bin sh
WORKDIR /beamlit

# Install the application dependencies.
COPY pyproject.toml /beamlit/pyproject.toml
COPY uv.lock /beamlit/uv.lock
RUN uv sync --no-cache

COPY README.md /beamlit/README.md
COPY LICENSE /beamlit/LICENSE
COPY src /beamlit/src

ENV PATH="/beamlit/.venv/bin:$PATH"

ENTRYPOINT [{cmd_str}]
"""

def generate_beamlit_deployment(directory: str):
    """
    Generates all necessary deployment files for Beamlit agents and functions.

    Args:
        directory (str): Target directory for generated files

    Creates:
        - Agent and function YAML configurations
        - Dockerfiles for each deployment
        - Directory structure for agents and functions
    """
    settings = init()
    logger = getLogger(__name__)
    logger.info(f"Importing server module: {settings.server.module}")
    functions: list[tuple[Resource, FunctionDeployment]] = []
    agents: list[tuple[Resource, AgentDeployment]] = []
    for agent in get_resources("agent"):
        agent_deployment = get_beamlit_deployment_from_resource(agent)
        if agent_deployment:
            agents.append((agent, agent_deployment))
    for function in get_resources("function"):
        function_deployment = get_beamlit_deployment_from_resource(function)
        if function_deployment:
            functions.append((function, function_deployment))

    agents_dir = os.path.join(directory, "agents")
    functions_dir = os.path.join(directory, "functions")
    # Create directory if it doesn't exist
    os.makedirs(agents_dir, exist_ok=True)
    os.makedirs(functions_dir, exist_ok=True)
    for (resource, agent) in agents:
        # write deployment file
        agent_dir = os.path.join(agents_dir, agent.agent)
        os.makedirs(agent_dir, exist_ok=True)
        with open(os.path.join(agent_dir, f"agent.yaml"), "w") as f:
            content = get_agent_yaml(agent, functions, settings)
            f.write(content)
        # write dockerfile for build
        with open(os.path.join(agent_dir, f"Dockerfile"), "w") as f:
            content = dockerfile("agent", resource, agent)
            f.write(content)
    for (resource, function) in functions:
        # write deployment file
        function_dir = os.path.join(functions_dir, function.function)
        os.makedirs(function_dir, exist_ok=True)
        with open(os.path.join(function_dir, f"function.yaml"), "w") as f:
            content = get_function_yaml(function, settings)
            f.write(content)
        # write dockerfile for build
        with open(os.path.join(function_dir, f"Dockerfile"), "w") as f:
            content = dockerfile("function", resource, function)
            f.write(content)


from logging import getLogger

from starlette.middleware.base import BaseHTTPMiddleware


class AccessLogMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        logger = getLogger(__name__)
        response = await call_next(request)
        process_time = response.headers.get("X-Process-Time")
        rid_header = response.headers.get("X-Request-Id")
        request_id = rid_header or response.headers.get("X-Beamlit-Request-Id")
        logger.info(f"{request.method} {request.url.path} {response.status_code} {process_time}ms rid={request_id}")
        return response



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.flavor import Flavor
    from ..models.function_kit import FunctionKit
    from ..models.pod_template_spec import PodTemplateSpec
    from ..models.runtime import Runtime
    from ..models.serverless_config import ServerlessConfig
    from ..models.spec_configuration import SpecConfiguration
    from ..models.store_function_parameter import StoreFunctionParameter


T = TypeVar("T", bound="FunctionSpec")


@_attrs_define
class FunctionSpec:
    """Function specification

    Attributes:
        configurations (Union[Unset, SpecConfiguration]): Agent configuration, this is a key value storage. In your
            agent you can retrieve the value with config[key]
        enabled (Union[Unset, bool]): Enable or disable the agent
        flavors (Union[Unset, List['Flavor']]): Types of hardware available for deployments
        integration_connections (Union[Unset, List[str]]):
        pod_template (Union[Unset, PodTemplateSpec]): Pod template specification
        policies (Union[Unset, List[str]]):
        runtime (Union[Unset, Runtime]): Set of configurations for a deployment
        serverless_config (Union[Unset, ServerlessConfig]): Configuration for a serverless deployment
        description (Union[Unset, str]): Function description, very important for the agent function to work with an LLM
        kit (Union[Unset, List['FunctionKit']]): The kit of the function deployment
        parameters (Union[Unset, List['StoreFunctionParameter']]): Function parameters, for your function to be callable
            with Agent
        store_id (Union[Unset, str]): Store id
    """

    configurations: Union[Unset, "SpecConfiguration"] = UNSET
    enabled: Union[Unset, bool] = UNSET
    flavors: Union[Unset, List["Flavor"]] = UNSET
    integration_connections: Union[Unset, List[str]] = UNSET
    pod_template: Union[Unset, "PodTemplateSpec"] = UNSET
    policies: Union[Unset, List[str]] = UNSET
    runtime: Union[Unset, "Runtime"] = UNSET
    serverless_config: Union[Unset, "ServerlessConfig"] = UNSET
    description: Union[Unset, str] = UNSET
    kit: Union[Unset, List["FunctionKit"]] = UNSET
    parameters: Union[Unset, List["StoreFunctionParameter"]] = UNSET
    store_id: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        configurations: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.configurations, Unset):
            configurations = self.configurations.to_dict()

        enabled = self.enabled

        flavors: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.flavors, Unset):
            flavors = []
            for componentsschemas_flavors_item_data in self.flavors:
                componentsschemas_flavors_item = (
                    componentsschemas_flavors_item_data.to_dict()
                )
                flavors.append(componentsschemas_flavors_item)

        integration_connections: Union[Unset, List[str]] = UNSET
        if not isinstance(self.integration_connections, Unset):
            integration_connections = self.integration_connections

        pod_template: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.pod_template, Unset):
            pod_template = self.pod_template.to_dict()

        policies: Union[Unset, List[str]] = UNSET
        if not isinstance(self.policies, Unset):
            policies = self.policies

        runtime: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.runtime, Unset):
            runtime = self.runtime.to_dict()

        serverless_config: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.serverless_config, Unset):
            serverless_config = self.serverless_config.to_dict()

        description = self.description

        kit: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.kit, Unset):
            kit = []
            for kit_item_data in self.kit:
                kit_item = kit_item_data.to_dict()
                kit.append(kit_item)

        parameters: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.parameters, Unset):
            parameters = []
            for parameters_item_data in self.parameters:
                parameters_item = parameters_item_data.to_dict()
                parameters.append(parameters_item)

        store_id = self.store_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if configurations is not UNSET:
            field_dict["configurations"] = configurations
        if enabled is not UNSET:
            field_dict["enabled"] = enabled
        if flavors is not UNSET:
            field_dict["flavors"] = flavors
        if integration_connections is not UNSET:
            field_dict["integrationConnections"] = integration_connections
        if pod_template is not UNSET:
            field_dict["podTemplate"] = pod_template
        if policies is not UNSET:
            field_dict["policies"] = policies
        if runtime is not UNSET:
            field_dict["runtime"] = runtime
        if serverless_config is not UNSET:
            field_dict["serverlessConfig"] = serverless_config
        if description is not UNSET:
            field_dict["description"] = description
        if kit is not UNSET:
            field_dict["kit"] = kit
        if parameters is not UNSET:
            field_dict["parameters"] = parameters
        if store_id is not UNSET:
            field_dict["store_id"] = store_id

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        from ..models.flavor import Flavor
        from ..models.function_kit import FunctionKit
        from ..models.pod_template_spec import PodTemplateSpec
        from ..models.runtime import Runtime
        from ..models.serverless_config import ServerlessConfig
        from ..models.spec_configuration import SpecConfiguration
        from ..models.store_function_parameter import StoreFunctionParameter

        d = src_dict.copy()
        _configurations = d.pop("configurations", UNSET)
        configurations: Union[Unset, SpecConfiguration]
        if isinstance(_configurations, Unset):
            configurations = UNSET
        else:
            configurations = SpecConfiguration.from_dict(_configurations)

        enabled = d.pop("enabled", UNSET)

        flavors = []
        _flavors = d.pop("flavors", UNSET)
        for componentsschemas_flavors_item_data in _flavors or []:
            componentsschemas_flavors_item = Flavor.from_dict(
                componentsschemas_flavors_item_data
            )

            flavors.append(componentsschemas_flavors_item)

        integration_connections = cast(
            List[str], d.pop("integrationConnections", UNSET)
        )

        _pod_template = d.pop("podTemplate", UNSET)
        pod_template: Union[Unset, PodTemplateSpec]
        if isinstance(_pod_template, Unset):
            pod_template = UNSET
        else:
            pod_template = PodTemplateSpec.from_dict(_pod_template)

        policies = cast(List[str], d.pop("policies", UNSET))

        _runtime = d.pop("runtime", UNSET)
        runtime: Union[Unset, Runtime]
        if isinstance(_runtime, Unset):
            runtime = UNSET
        else:
            runtime = Runtime.from_dict(_runtime)

        _serverless_config = d.pop("serverlessConfig", UNSET)
        serverless_config: Union[Unset, ServerlessConfig]
        if isinstance(_serverless_config, Unset):
            serverless_config = UNSET
        else:
            serverless_config = ServerlessConfig.from_dict(_serverless_config)

        description = d.pop("description", UNSET)

        kit = []
        _kit = d.pop("kit", UNSET)
        for kit_item_data in _kit or []:
            kit_item = FunctionKit.from_dict(kit_item_data)

            kit.append(kit_item)

        parameters = []
        _parameters = d.pop("parameters", UNSET)
        for parameters_item_data in _parameters or []:
            parameters_item = StoreFunctionParameter.from_dict(parameters_item_data)

            parameters.append(parameters_item)

        store_id = d.pop("store_id", UNSET)

        function_spec = cls(
            configurations=configurations,
            enabled=enabled,
            flavors=flavors,
            integration_connections=integration_connections,
            pod_template=pod_template,
            policies=policies,
            runtime=runtime,
            serverless_config=serverless_config,
            description=description,
            kit=kit,
            parameters=parameters,
            store_id=store_id,
        )

        function_spec.additional_properties = d
        return function_spec

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.agent_deployment_history import AgentDeploymentHistory
from ...types import Response


def _get_kwargs(
    request_id: str,
) -> dict[str, Any]:
    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": f"/history/agents/{request_id}",
    }

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[list["AgentDeploymentHistory"]]:
    if response.status_code == 200:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = AgentDeploymentHistory.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[list["AgentDeploymentHistory"]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    request_id: str,
    *,
    client: AuthenticatedClient,
) -> Response[list["AgentDeploymentHistory"]]:
    """Get all history for a specific request ID from all agents

    Args:
        request_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[list['AgentDeploymentHistory']]
    """

    kwargs = _get_kwargs(
        request_id=request_id,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    request_id: str,
    *,
    client: AuthenticatedClient,
) -> Optional[list["AgentDeploymentHistory"]]:
    """Get all history for a specific request ID from all agents

    Args:
        request_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        list['AgentDeploymentHistory']
    """

    return sync_detailed(
        request_id=request_id,
        client=client,
    ).parsed


async def asyncio_detailed(
    request_id: str,
    *,
    client: AuthenticatedClient,
) -> Response[list["AgentDeploymentHistory"]]:
    """Get all history for a specific request ID from all agents

    Args:
        request_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[list['AgentDeploymentHistory']]
    """

    kwargs = _get_kwargs(
        request_id=request_id,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    request_id: str,
    *,
    client: AuthenticatedClient,
) -> Optional[list["AgentDeploymentHistory"]]:
    """Get all history for a specific request ID from all agents

    Args:
        request_id (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        list['AgentDeploymentHistory']
    """

    return (
        await asyncio_detailed(
            request_id=request_id,
            client=client,
        )
    ).parsed



from typing import Any, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="RuntimeType0")


@_attrs_define
class RuntimeType0:
    """Set of configurations for a deployment

    Attributes:
        args (Union[Unset, List[Any]]): The arguments to pass to the deployment runtime
        command (Union[Unset, List[Any]]): The command to run the deployment
        envs (Union[Unset, List[Any]]): The environment variables to set in the deployment. Should be a list of
            Kubernetes EnvVar types
        image (Union[Unset, str]): The Docker image for the deployment
        model (Union[Unset, str]): The slug name of the origin model. Only used if the deployment is a ModelDeployment
        type (Union[Unset, str]): The type of origin for the deployment
    """

    args: Union[Unset, List[Any]] = UNSET
    command: Union[Unset, List[Any]] = UNSET
    envs: Union[Unset, List[Any]] = UNSET
    image: Union[Unset, str] = UNSET
    model: Union[Unset, str] = UNSET
    type: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        args: Union[Unset, List[Any]] = UNSET
        if not isinstance(self.args, Unset):
            args = self.args

        command: Union[Unset, List[Any]] = UNSET
        if not isinstance(self.command, Unset):
            command = self.command

        envs: Union[Unset, List[Any]] = UNSET
        if not isinstance(self.envs, Unset):
            envs = self.envs

        image = self.image

        model = self.model

        type = self.type

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if args is not UNSET:
            field_dict["args"] = args
        if command is not UNSET:
            field_dict["command"] = command
        if envs is not UNSET:
            field_dict["envs"] = envs
        if image is not UNSET:
            field_dict["image"] = image
        if model is not UNSET:
            field_dict["model"] = model
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        args = cast(List[Any], d.pop("args", UNSET))

        command = cast(List[Any], d.pop("command", UNSET))

        envs = cast(List[Any], d.pop("envs", UNSET))

        image = d.pop("image", UNSET)

        model = d.pop("model", UNSET)

        type = d.pop("type", UNSET)

        runtime_type_0 = cls(
            args=args,
            command=command,
            envs=envs,
            image=image,
            model=model,
            type=type,
        )

        runtime_type_0.additional_properties = d
        return runtime_type_0

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.runtime_readiness_probe import RuntimeReadinessProbe
    from ..models.runtime_resources import RuntimeResources


T = TypeVar("T", bound="Runtime")


@_attrs_define
class Runtime:
    """Set of configurations for a deployment

    Attributes:
        args (Union[Unset, list[Any]]): The arguments to pass to the deployment runtime
        command (Union[Unset, list[Any]]): The command to run the deployment
        envs (Union[Unset, list[Any]]): The environment variables to set in the deployment. Should be a list of
            Kubernetes EnvVar types
        image (Union[Unset, str]): The Docker image for the deployment
        model (Union[Unset, str]): The slug name of the origin model. Only used if the deployment is a ModelDeployment
        readiness_probe (Union[Unset, RuntimeReadinessProbe]): The readiness probe. Should be a Kubernetes Probe type
        resources (Union[Unset, RuntimeResources]): The resources for the deployment. Should be a Kubernetes
            ResourceRequirements type
        type_ (Union[Unset, str]): The type of origin for the deployment
    """

    args: Union[Unset, list[Any]] = UNSET
    command: Union[Unset, list[Any]] = UNSET
    envs: Union[Unset, list[Any]] = UNSET
    image: Union[Unset, str] = UNSET
    model: Union[Unset, str] = UNSET
    readiness_probe: Union[Unset, "RuntimeReadinessProbe"] = UNSET
    resources: Union[Unset, "RuntimeResources"] = UNSET
    type_: Union[Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        args: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.args, Unset):
            args = self.args

        command: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.command, Unset):
            command = self.command

        envs: Union[Unset, list[Any]] = UNSET
        if not isinstance(self.envs, Unset):
            envs = self.envs

        image = self.image

        model = self.model

        readiness_probe: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.readiness_probe, Unset):
            readiness_probe = self.readiness_probe.to_dict()

        resources: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.resources, Unset):
            resources = self.resources.to_dict()

        type_ = self.type_

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if args is not UNSET:
            field_dict["args"] = args
        if command is not UNSET:
            field_dict["command"] = command
        if envs is not UNSET:
            field_dict["envs"] = envs
        if image is not UNSET:
            field_dict["image"] = image
        if model is not UNSET:
            field_dict["model"] = model
        if readiness_probe is not UNSET:
            field_dict["readiness_probe"] = readiness_probe
        if resources is not UNSET:
            field_dict["resources"] = resources
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.runtime_readiness_probe import RuntimeReadinessProbe
        from ..models.runtime_resources import RuntimeResources

        if not src_dict:
            return None
        d = src_dict.copy()
        args = cast(list[Any], d.pop("args", UNSET))

        command = cast(list[Any], d.pop("command", UNSET))

        envs = cast(list[Any], d.pop("envs", UNSET))

        image = d.pop("image", UNSET)

        model = d.pop("model", UNSET)

        _readiness_probe = d.pop("readiness_probe", UNSET)
        readiness_probe: Union[Unset, RuntimeReadinessProbe]
        if isinstance(_readiness_probe, Unset):
            readiness_probe = UNSET
        else:
            readiness_probe = RuntimeReadinessProbe.from_dict(_readiness_probe)

        _resources = d.pop("resources", UNSET)
        resources: Union[Unset, RuntimeResources]
        if isinstance(_resources, Unset):
            resources = UNSET
        else:
            resources = RuntimeResources.from_dict(_resources)

        type_ = d.pop("type", UNSET)

        runtime = cls(
            args=args,
            command=command,
            envs=envs,
            image=image,
            model=model,
            readiness_probe=readiness_probe,
            resources=resources,
            type_=type_,
        )

        runtime.additional_properties = d
        return runtime

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
